// Helpful Tip - I use an indent size of 3 because I like it.  2 is to little, and 4 is too much.  If you happen to use a 
// different size, you can fixup this file to match your new editor settings using one of two methods:
//
// 1. Hit Ctrl-A to select the files and use Edit | Advanced | Format Selection
// 2. Remove the last } and type it again.  Whenever you type a }, everything in the enclosing space is reformatted.

using System;
using System.Runtime.InteropServices;
using System.Windows.Forms;

#region Clarion basic datatypes that exactly match to .NET datatypes
using CWByte     = System.Byte;
using CWShort    = System.Int16;
using CWUShort   = System.UInt16;
using CWLong     = System.Int32;
using CWULong    = System.UInt32;
using CWSigned   = System.Int32;
using CWUnsigned = System.UInt32;
using CWSReal    = System.Single;
using CWReal     = System.Double;
#endregion

namespace CWInterop
{
   public class CWLink
   {
      #region External Interface for SharedDLL Clarion DLL
      [MarshalAs(UnmanagedType.BStr)]
      static Type Zx = typeof(System.String);
      
      [DllImport("SharedDLL", EntryPoint="PUBLICPROCEDURE")]
      public static extern void PublicProcedure();

      [DllImport("SharedDLL", EntryPoint="TAKESLONGRETURNSLONG")]
      public static extern CWLong TakesLongReturnsLong(CWLong value);

      [DllImport("SharedDLL", EntryPoint="TAKESSTRING")]
      public static extern void TakesString([MarshalAs(UnmanagedType.BStr)] string value);

      [DllImport("SharedDLL", EntryPoint="TAKESANDRETURNSSTRING")]
      [return: MarshalAs(CWBString)]
      public static extern string TakesAndReturnsString([MarshalAs(CWBString)] string value);

      [DllImport("SharedDLL", EntryPoint="TAKESANDRETURNSDATE")]
      public static extern CWDateLong TakesAndReturnsDate_External(CWDateLong value);
      public static DateTime TakesAndReturnsDate(DateTime value) 
      {
         return (DateTime) TakesAndReturnsDate_External(value);
      }

      [DllImport("SharedDLL", EntryPoint="TAKESANDRETURNSTIME")]
      public static extern CWTimeLong TakesAndReturnsTime_External(CWTimeLong value);
      public static DateTime TakesAndReturnsTime(DateTime value) 
      {
         return (DateTime) TakesAndReturnsTime_External(value);
      }

      /// <summary>
      /// Take and return a .NET <c>decimal</c> value from Clarion using a <c>bstring/c> as the intermediate representation.
      /// This demonstrates a basic wrapper-method technique that can solve quite a lot of difficult data marshaling problems
      /// relatively painlessly.  The <c>_wrapper</c> method passes a bstring back and forth to Clarion.  The .NET wrapper does
      /// the conversion to and from a string using native .NET methods.
      /// </summary>
      [DllImport("SharedDLL", EntryPoint = "TAKESANDRETURNSDECIMAL")]
      [return: MarshalAs(CWBString)]
      private static extern string TakesAndReturnsDecimal_Wrapper([MarshalAs(CWBString)] string value);
      public static decimal TakesAndReturnsDecimal(decimal value)
      {
         return Decimal.Parse(TakesAndReturnsDecimal_Wrapper(value.ToString()));
      }
      #endregion

      #region Clarion Datatype Converters
      // Shortcut for Clarion BSTRING to save typing
      const UnmanagedType CWBString = UnmanagedType.BStr;

      /// <summary>
      ///    This structure converts back and forth between a .NET DateTime object and a Clarion DATE.  For
      ///    details see http://www.clarionmag.com.  This uses implicit conversion to convert the DATE portion 
      ///    of a DateTime to the 32 bit integer equivalent used by Clarion.   This class struct should 
      ///    never be created explicitly.  It only exists to marshal data betwen Clarion and .NET.
      /// </summary>
      public struct CWDateLong
      {
         // Date portion of DateTime as a Clarion LONG
         public Int32 _cwDateAsLong;

         // Offset in ticks between .NET start date and Clarion start date.
         const long DATEOFFSET = 657432;
   
         // Implicit operator to convert the Int32 returned by a Clarion function to a DateTime object.
         public static implicit operator DateTime(CWDateLong obj)
         {
            return new DateTime((obj._cwDateAsLong + DATEOFFSET) * TimeSpan.TicksPerDay);
         }

         // Implicit operator to convert a DateTime to the 32 bit Clarion value
         public static implicit operator CWDateLong(DateTime value)
         {
            CWDateLong result = new CWDateLong();
            result._cwDateAsLong = Convert.ToInt32((value.Ticks / TimeSpan.TicksPerDay) - DATEOFFSET);
            return result;
         }
      }

      /// <summary>
      ///    This structure converts back and forth between a .NET DateTime object and a Clarion TIME.  For
      ///    details see http://www.clarionmag.com.  This uses implicit conversion to convert the TIME portion 
      ///    of a DateTime to the 32 bit integer equivalent used by Clarion.   This class struct should 
      ///    never be created explicitly.  It only exists to marshal data betwen Clarion and .NET.
      /// </summary>
      public struct CWTimeLong
      {
         // Time portion of DateTime as a Clarion LONG
         private Int32 _cwTimeAsLong;
   
         // Implicit operator to convert the Int32 returned by a Clarion function to a DateTime object.
         public static implicit operator DateTime(CWTimeLong obj)
         {
            return new DateTime() + new TimeSpan((obj._cwTimeAsLong - 1) * TimeSpan.TicksPerSecond / 100);
         }

         // Implicit operator to convert a DateTime to the 32 bit Clarion value
         public static implicit operator CWTimeLong(DateTime value)
         {
            CWTimeLong result = new CWTimeLong();
            result._cwTimeAsLong = Convert.ToInt32(value.Millisecond / 10 + value.Second * 100 
               + value.Minute * 6000 + value.Hour * 360000 + 1);
            return result;
         }
      }

      /// <summary>
      /// This converter is similar to the date converter shown above.  I'm using a brute-force representation of a string for
      /// the decimal, because it's very portable and it performs well enough for just about any purpose.  It is possible to
      /// pass the .NET decimal value to Clarion and parse it directly to a Clarion decimal, but it would be a lot of work for
      /// no real benefit.  I leave it here for educational purposes, but I tend to use the simpler double-wrapper shown
      /// above, which is easier to consume in .NET.
      /// </summary>
      public struct CWDecimal
      {
         [MarshalAs(UnmanagedType.BStr)]
         public string _decimalAsString;

         public static implicit operator CWDecimal(decimal value)
         {
            CWDecimal result = new CWDecimal();
            result._decimalAsString = value.ToString();
            return result;
         }

         public static implicit operator decimal(CWDecimal value)
         {
            return Decimal.Parse(value._decimalAsString);
         }
      }
      #endregion
   }
}